// meric superkondenzatoru
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <EEPROM.h>

// nastavena adresa 0x3F a poradi pinu displeje na prevodniku: en,rw,rs,d4,d5,d6,d7,bl,blpol
LiquidCrystal_I2C lcd(0x3F, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);

const int ODPOR_VNITRNICH_VODICU = 6000; //v mikroohmech

const char *mainMenu[7];
const char * nastaveni[13];
unsigned int params[12];
int mainMenuN = 6;
int nastaveniN = 12;
int posOfSave = 12;
unsigned long lastClick = 0;
int oldNav = 2;
int oldProgress = -1;
unsigned long oldProgressTime = 0;
unsigned long voltageTime = 0;

volatile boolean esc = false;

void setup() {
  byte g0[8] = {B00000, B00000, B00000, B10101, B10000, B10000, B10000, B10101}; //0
  byte g1[8] = {B00000, B00000, B00000, B10101, B10100, B10100, B10100, B10101}; //1
  byte g2[8] = {B00000, B00000, B00000, B10101, B10101, B10101, B10101, B10101}; //2
  byte g3[8] = {B00000, B00000, B00000, B10101, B00000, B00000, B00000, B10101}; //3
  byte g4[8] = {B00000, B00000, B00000, B10101, B00001, B00001, B00001, B10101}; //4
  byte g5[8] = {B00000, B00000, B00000, B10101, B10001, B10001, B10001, B10101}; //5
  byte g6[8] = {B00000, B00000, B00001, B00001, B01001, B11111, B01000, B00000}; //6
  byte g7[8] = {B10101, B10101, B10101, B10101, B10101, B10101, B10101, B10101}; //7

  pinMode(7, OUTPUT); digitalWrite(7, HIGH);       // vybijeni 5A - PD7
  pinMode(6, OUTPUT); digitalWrite(6, HIGH);       // vybijeni 0.5A - PD6
  pinMode(5, OUTPUT); digitalWrite(5, HIGH);       // vybijeni 50 mA - PD5
  pinMode(4, OUTPUT); digitalWrite(4, LOW);        // nabijeni 5A - PD4
  pinMode(3, OUTPUT); digitalWrite(3, LOW);        // nabijeni 0.5A - PD3
  pinMode(13, OUTPUT);digitalWrite(13, LOW);       // nabijeni 0.05A - PB5
  pinMode(8, OUTPUT); digitalWrite(8, HIGH);       // vetrak on
  pinMode(9, OUTPUT);                              // zvuk - PB1
  //pinMode(9, OUTPUT);                            // ventilator - PD3
  pinMode(2, INPUT_PULLUP);                        // TL ESC pres interrupt - INT0
  lcd.begin(20, 4);                                // zavedeni grafiky do RAM displeje
  lcd.createChar(0, g0); lcd.createChar(1, g1); lcd.createChar(2, g2); lcd.createChar(3, g3);
  lcd.createChar(4, g4); lcd.createChar(5, g5); lcd.createChar(6, g6); lcd.createChar(7, g7);
  analogReference(EXTERNAL);                       // reference 2,5 V z TL431
  lcd.clear();
  lcd.print("       Meric");                       // vypis hlavicky
  lcd.setCursor(0, 1);
  lcd.print("  superkondenzatoru");                // vypis hlavicky
  lcd.setCursor(0, 3);
  lcd.print("     verze 2.1");
  attachInterrupt(0, escHandle, FALLING);// interrupt
  delay(3000);

  mainMenu[0] = "Nabit kondenzator";
  mainMenu[1] = "Vybit kondenzator";
  mainMenu[2] = "Zmerit kapacitu";
  mainMenu[3] = "Zmerit ESR";
  mainMenu[4] = "Formovani";
  mainMenu[5] = "Svod";
  mainMenu[6] = "Nastaveni";

  nastaveni[0] = "Nabijeni I:";
  nastaveni[1] = "Nabijeni U:";
  nastaveni[2] = "Vybijeni I:";
  nastaveni[3] = "Vybijeni U:";
  nastaveni[4] = "Mereni C U1:";
  nastaveni[5] = "Mereni C U2:";
  nastaveni[6] = "Mer. ESR U1:";
  nastaveni[7] = "Mer. ESR U2:";
  nastaveni[8] = "Formovani t:";
  nastaveni[9] = "Svod t1:";
  nastaveni[10] = "Svod t2:";
  nastaveni[11] = "Zvuky:";
  nastaveni[12] = "Ulozit do EEPROM";

  EEPROM.get(0, params);
  if (params[0] == 0) {
    params[0] = 2; //nab I
    params[1] = 25; //nab U
    params[2] = 2; //vyb I
    params[3] = 5; //vyb U
    params[4] = 20; //mer C u1
    params[5] = 10; //mer C u2
    params[6] = 20; //esr u1
    params[7] = 10; //esr u2
    params[8] = 60; //formovani
    params[9] = 5; //svod t1 - prodleva
    params[10] = 60; //svod t2 - cas
    params[11] = 1; //zvuk
    //save pos
  }

  
  digitalWrite(8, LOW); //vetrak off
  lcd.clear();
}

void escHandle() {
  esc = true;
}

unsigned int getParam(int index) { //mapovani
  switch (index) {
    case 0: return params[0]; //nab U
    case 1: return params[1]; //nab I
    case 2: return params[2]; //vyb I
    case 3: return params[3]; //vyb U
    case 4: return params[4]; //C od
    case 5: return params[5]; //C do
    case 6: return params[6]; //ESR 1
    case 7: return params[7]; //ESR 2
    case 8: return params[8]; //form 1
    case 9: return params[9]; //svod 1 - prodleva
    case 10: return params[10]; //svod 2 - cas
    case 11: return params[11]; //zvuk
  }
}

void setParam(int index, unsigned value) { //mapovani
  switch (index) {
    case 0: params[0] = value; break;
    case 1: params[1] = value; break;
    case 2: params[2] = value; break;
    case 3: params[3] = value; break;
    case 4: params[4] = value; break;
    case 5: params[5] = value; break;
    case 6: params[6] = value; break;
    case 7: params[7] = value; break;
    case 8: params[8] = value; break;
    case 9: params[9] = value; break;
    case 10: params[10] = value; break;
    case 11: params[11] = value; break;
  }
}

/*
 * isXXX = convenience metody ... index bude podle getParam()
 */

int isVoltage(int index) { //napeti
  switch (index) {
    case 1: return 1; //nabijeci
    case 3: return 2; //vybijeci
    case 4: return 3; //mereni C od
    case 5: return 4; //mereni C do
    case 6: return 5; //ESR od
    case 7: return 6; //ESR do
    default: return 0; //neni napeti
  }
}
boolean isCurrent(int index) { //proud
  return index == 0 || index == 2;
}
int isTime(int index) {
  if (index == 8 || index == 10) { //velky casy
    return 1;
  } else if (index == 9) { //kratky cas
    return 2;
  } else {
    return 0;
  }
}
boolean isSound(int index) {
  return index == 11;
}

boolean soundEnabled() {
  return getParam(11);
}

void initProgressBar() {
  oldProgress = 0;
  oldProgressTime = 0;
  lcd.setCursor(0, 3);
  lcd.write((byte)0);
  for (int i = 1; i < 19; i++) {
    lcd.write(3);
  }
  lcd.write(4);
  lcd.setCursor(0, 3);
}


void updateProgress(int cislo) {
  unsigned long currentMillis = millis();
  if (currentMillis - oldProgressTime < 10) { // refresh
    delay(1);
    return;
  }
  oldProgressTime = currentMillis;

  cislo = min(cislo, 59);  cislo = max(cislo, 0);

  if (oldProgress >= cislo) {
    delay(1);
    return;
  }

  //oldProgress < cislo
  oldProgress++;
  lcd.setCursor(oldProgress / 3, 3);
  if (oldProgress < 57) {
    lcd.write(oldProgress % 3);
  } else if (oldProgress == 57) {
    lcd.write(5);
  }  else if (oldProgress == 58) {
    lcd.write(2);
  }
}

void showVoltage(){
  unsigned long currentMillis = millis();
  if (currentMillis - voltageTime < 1000) { // refresh
    return;
  }
  voltageTime = currentMillis;

  float u = (analogRead(2) * 2.5f) / 1023.0f;

  lcd.setCursor(14,0);
  lcd.print(u); lcd.print("V");
}

void printTimePart(int val) {
  if (val < 10) {
    lcd.print("0");
  }
  lcd.print(val);
}

void printValue(unsigned int value, int index, int col, int row) {
  if (isCurrent(index)) { //Proud - Enum
    lcd.setCursor(col, row);
    switch (value) {
      case 1: lcd.print("50mA"); break;
      case 2: lcd.print("0.5A"); break;
      case 3: lcd.print("5.0A"); break;
    }
  } else if (isTime(index)) { //Time
    lcd.setCursor(col - 1, row);
    printTimePart(value / 60);
    lcd.print(":");
    printTimePart(value % 60);
  } else if (isSound(index)) { // Zvuk
    lcd.setCursor(col, row);
    if (value) {
      lcd.print("ANO");
    } else {
      lcd.print("NE ");
    }
  } else if (isVoltage(index)) { // Napeti
    lcd.setCursor(col, row);
    lcd.print(value / 10);
    lcd.print(".");
    lcd.print(value % 10);
    lcd.print("V");
  } else { //jen to vypis
    lcd.setCursor(col, row);
    lcd.print(value);
  }

}

void drawOptions(const char **opt, int Nopt, int pos, boolean edit) {
  int from = pos - 1; //odkud vykreslovat
  int to = pos + 1; // kam vykreslovat
  if (from < 0) { //opravit pokud je to prvni item
    from = 0;
    to++;
  }
  if (to > Nopt) { //opravit pokud je to posledni item
    to = Nopt;
    from--;
  }

  int row = 1;
  for (int i = from; i <= to; i++, row++) { //vykresli jednotlive polozky
    lcd.setCursor(0, row);
    if (pos == i) { //vykresli sipku kde ma byt
      lcd.write(126);
    } else {
      lcd.write(" ");
    }
    lcd.write(" ");
    lcd.write(opt[i]);

    if (edit && (i != posOfSave)) { //pokud je edit, je to SETUP(index 11)... a posledni je bez hodnoty
      printValue(getParam(i), i, 16, row);
    }
  }
}

int getKey() {
  while (millis() - lastClick < 300) {}
  while (1) {
    int key1 = analogRead(3);
    delay(1);
    int key2 = analogRead(3);
    if (abs(key1 - key2) < 5) {
      lastClick = millis();
      if (key1 < 150) { //enter
        while(analogRead(3)<750){} //dokud nepusti enter tak nic nebude
        return 0;
      } else if (key1 < 450) { //up
        return 1;
      } else if (key1 < 750) { //down
        return 2;
      }
    }
    if (esc) {
      esc = false;
      return 3;
    }
  }
}

unsigned int changeValue(unsigned int value, int index, bool up) {
  if (isTime(index) == 1) { //velky cas
    if (up) {
      value += 10;
    } else {
      value -= 10;
    };
    value = min(value, 5990);
    value = max(value, 10);
  } else {
    if (up) { //vsechno ostatni jde po 1
      value++;
    } else {
      value--;
    }
    if (isTime(index) == 2) { //kratky cas
      value = min(value, 60);
      value = max(value, 1);
    }

    if (isCurrent(index)) { //Proud
      if (value > 3) value = 1;
      if (value < 1) value = 3;
    }

    if (isSound(index)) { //zvuk
      value %= 2;
    }

    if (isVoltage(index) == 1) { //nab U
      value = min(value, 25);
      value = max(value, 10);
    } else if (isVoltage(index) == 2) { //vyb U
      value = min(value, 25);
      value = max(value, 1);
    } else if (isVoltage(index) == 3) { // Mereni C od
      if (value <= getParam(5)) {
        value++;
        if (soundEnabled())tone(9, 200, 400);
      }
      value = min(value, 22);
      value = max(value, 9);
    } else if (isVoltage(index) == 4) { //Mereni C do
      if (value >= getParam(4)) {
        value--;
        if (soundEnabled())tone(9, 200, 400);
      }
      value = min(value, 22);
      value = max(value, 9);
    } else if (isVoltage(index) == 5) { //ESR od
      if (value <= getParam(7)) {
        value++;
        if (soundEnabled())tone(9, 200, 400);
      }
      value = min(value, 24);
      value = max(value, 9);
    } else if (isVoltage(index) == 6) { //ESR do
      if (value >= getParam(6)) {
        value--;
        if (soundEnabled())tone(9, 200, 400);
      }
      value = min(value, 24);
      value = max(value, 9);
    }
  }
  return value;
}

//2 mody = mainMenu a settings(edit=true)
int menu(const char ** opt, int Nopt, const char * title, int nav, boolean edit) {
  do {
    lcd.clear(); //clear screen
    lcd.write(title);
    drawOptions(opt, Nopt, nav, edit);

    switch (getKey()) {
      case 1: { // nahoru
          nav--;
          if (nav < 0) nav = 0;
          break;
        }
      case 2: { // dolu
          nav++;
          if (nav > Nopt) nav = Nopt;
          break;
        }
      case 0: { // enter
          if (!edit)return nav;
          //Obsluha nastaveni
          if (nav == posOfSave) { //kliknul na save
            return nav;
          }
          int row;
          if (nav == 0) {
            row = 1;
          } else if (nav == Nopt) {
            row = 3;
          } else {
            row = 2;
          }

          lcd.setCursor(0, row);
          lcd.write(" ");

          if (isTime(nav)) { //cas se chova jinak
            lcd.setCursor(14, row);
          } else {
            lcd.setCursor(15, row);
          }
          lcd.write(126);

          unsigned int value = getParam(nav);
          do {
            int key = getKey();
            if (key == 0) { //enter - potvrd
              setParam(nav, value);
              break;
            } else if (key == 1) { // up
              value = changeValue(value, nav, 1);
            } else if (key == 2) { //down
              value = changeValue(value, nav, 0);
            } else if (key == 3) { //esc - cancel
              break;
            }
            printValue(value, nav, 16, row);
          } while (1);
          break;
        }
      case 3: { //esc
          return nav + 100;
          break;
        }
    }

  } while (1);
}

void vypni() {
  digitalWrite(7, HIGH);       // vybijeni 5A - PD7
  digitalWrite(6, HIGH);       // vybijeni 0.5A - PD6
  digitalWrite(5, HIGH);       // vybijeni 50 mA - PD5
  digitalWrite(4, LOW);        // nabijeni 5A - PD4
  digitalWrite(3, LOW);        // nabijeni 0.5A - PD3
  digitalWrite(13, LOW);      // nabijeni 0.05A - PB5
}

void done(bool error, int type) {
  lcd.setCursor(0, 3);
  if (type == 0) {
    lcd.print(" ZRUSENO!           ");
  } else if (type == 1) {
    lcd.print(" HOTOVO!            ");
  } else if (type == 2) {
    lcd.print(" VYBITO! (          ");

    int r = (analogRead(2) * 25) / 1023;
    printValue(r, 1, 11, 3);
    lcd.print(")");
  } else if (type == 3) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Prilis velke napeti");
    lcd.setCursor(0, 1);
    lcd.print("na kondenzatoru.");
    lcd.setCursor(0, 2);
    lcd.print("(");
    int r = (analogRead(2) * 25) / 1023;
    if (r == 1023) {
      lcd.print(">2.5V");
    } else {
      printValue(r, 1, 1, 2);
    }
    lcd.print(")");
  } else if (type == 4) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Prilis velky ubytek");
    lcd.setCursor(0, 1);
    lcd.print("napeti pred merenim.");
    lcd.setCursor(0, 2);
    lcd.print("Je-li mozno, snizte");
    lcd.setCursor(0, 3);
    lcd.print("merici proud.");
  } else if (type == 5) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Prekroceni doby");
    lcd.setCursor(0, 1);
    lcd.print("mereni (20h).");
    lcd.setCursor(0, 2);
    lcd.print("Je-li mozno, zvyste");
    lcd.setCursor(0, 3);
    lcd.print("merici proud.");
  } else if (type == 6) {
    lcd.clear();
    lcd.setCursor(0, 1);
    lcd.print("Byl prekrocen limit");
    lcd.setCursor(0, 2);
    lcd.print("mereni ESR (5"); lcd.write(244); lcd.print(").");
  } else if (type == 7) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Prilis mala kapacita");
    lcd.setCursor(0, 1);
    lcd.print("kondenzatoru nebo");
    lcd.setCursor(0, 2);
    lcd.print("prilis velky proud");
    lcd.setCursor(0, 3);
    lcd.print("vybijeni.");
  } else {
    lcd.clear();
    lcd.setCursor(0, 1);
    lcd.print("Neznama chyba: "); lcd.print(type);
  }
  if (soundEnabled()) {
    for (int i = 0; i < 4; i++) {
      tone(9, error ? 200 : 1000, 100);
      delay(300);
    }
  }
  getKey();
}

int nabij(bool useProgressBar) {
  int endU = (getParam(1) * 1023) / 25;
  int currU = analogRead(2);
  int firstRead = currU;
  if (currU > endU) {
    return 0;
  }

  boolean success = true;
  initProgressBar();
  do { //kontroluj stav
    switch (getParam(0)) {//zacni nabijet
      case 1: { //50mA
          digitalWrite(13, HIGH);
          break;
        }
      case 2: { //0.5A
          digitalWrite(3, HIGH);
          break;
        }
      case 3: { //5.0A
          digitalWrite(4, HIGH);
          digitalWrite(8, HIGH); // bude zahul - bude vitr
          if (currU > 410) { //nabijis 5A a napeti je 1V+
            digitalWrite(3, HIGH); // zapni 0.5A zdroj
          }
          break;
        }
    }

    if (useProgressBar) {
      int val = map(currU, firstRead, endU, 0, 60);
      updateProgress(val);
    }

    if (esc) { //ted hned to ukonci
      esc = false;
      success = false;
      break;
    }
    delay(40);
    vypni();
    delay(10);
    currU = analogRead(2);
  } while (currU < endU); // dokud neni nabito
  vypni();
  digitalWrite(8, LOW); // vetrak off

  if (success) {
    return 1;
  }  else {
    return 2;
  }
}

int vybij(int toU = getParam(3)) {
  int endU = (toU * 1023) / 25;
  int currU = analogRead(2);
  int firstRead = currU;
  if (currU < endU) {
    return 0;
  }

  boolean success = true;
  initProgressBar();
  do { //kontroluj stav
    switch (getParam(2)) {//zacni vybijet
      case 1: { //50mA
          digitalWrite(5, LOW);
          break;
        }
      case 2: { //0.5A
          digitalWrite(6, LOW);
          break;
        }
      case 3: { //5.0A
          digitalWrite(7, LOW);
          digitalWrite(8, HIGH); // bude zahul - bude vitr
          break;
        }
    }

    int val = map(currU, firstRead, endU, 0, 59);
    updateProgress(val);

    if (esc) { //ted hned to ukonci
      esc = false;
      success = false;
      break;
    }
    delay(40);
    vypni();
    delay(5);
    currU = analogRead(2);
  } while (currU > endU);
  vypni();
  digitalWrite(8, LOW); // vetrak off

  if (success) {
    return 1;
  }  else {
    return 2;
  }
}

unsigned long getESR() {
  int usedI = 0;
  delay(1000);
  switch (getParam(2)) {//zacni vybijet
    case 1: { //50mA
        usedI = 5;
        digitalWrite(5, LOW);
        break;
      }
    case 2: { //0.5A
        usedI = 50;
        digitalWrite(6, LOW);
        break;
      }
    case 3: { //5.0A
        usedI = 500;
        digitalWrite(7, LOW);
        break;
      }
  }
  delay(2);

  int where = 1;
  if (analogRead(1) > 1020) {
    where = 0;
    if (analogRead(0) > 1020) {
      vypni();
      return 0;
    }
  }
  vypni();

  int after = 0;
  int before = 0;
  unsigned long avg = 0;
  for (int i = 0; i < 10; i++) {
    delay(100);

    before = analogRead(where);

    switch (getParam(2)) {//zacni vybijet
      case 1: { //50mA
          digitalWrite(5, LOW);
          break;
        }
      case 2: { //0.5A
          digitalWrite(6, LOW);
          break;
        }
      case 3: { //5.0A
          digitalWrite(7, LOW);
          break;
        }
    }
    delay(2);
    after = analogRead(where);
    vypni();
    avg += after - before;
  }
  avg /= 10;
  if (avg == 0) avg = 1;

  unsigned long result;
  if (where == 1) {
    result = (2500000 * avg) / 102300;
  } else {
    result = (2500000 * avg) / 10230;
  }
  result *= 100;
  result /= usedI;


  if (result > ODPOR_VNITRNICH_VODICU) {
    return result - ODPOR_VNITRNICH_VODICU;
  } else {
    return 1;
  }
}

void printESR(unsigned long number) {
  int d = (number == 0 ? 1 : ((int)(log10(number) + 1)));
  switch (d) {
    case 9: {
        number = roundIt(number, 1000000);
        String num = String(number);
        lcd.print(num.substring(0, 3));
        /*  lcd.print(".");
          lcd.print(num.substring(3, 4));*/
        lcd.print(""); lcd.write(244);
        break;
      }
    case 8: {
        number = roundIt(number, 100000);
        String num = String(number);
        lcd.print(num.substring(0, 2));
        lcd.print(".");
        lcd.print(num.substring(2, 3));
        lcd.print(""); lcd.write(244);
        break;
      }
    case 7: {
        number = roundIt(number, 10000);
        String num = String(number);
        lcd.print(num.substring(0, 1));
        lcd.print(".");
        lcd.print(num.substring(1, 3));
        lcd.print(""); lcd.write(244);
        break;
      }
    case 6: {
        number = roundIt(number, 1000);
        String num = String(number);
        lcd.print(num.substring(0, 3));
        /*  lcd.print(".");
          lcd.print(num.substring(3, 4));*/
        lcd.print("m"); lcd.write(244);
        break;
      }
    case 5: {
        number = roundIt(number, 100);
        String num = String(number);
        lcd.print(num.substring(0, 2));
        lcd.print(".");
        lcd.print(num.substring(2, 3));
        lcd.print("m"); lcd.write(244);
        break;
      }
    case 4: {
        number = roundIt(number, 10);
        String num = String(number);
        lcd.print(num.substring(0, 1));
        lcd.print(".");
        lcd.print(num.substring(1, 3));
        lcd.print("m"); lcd.write(244);
        break;
      }
    case 3: {
        String num = String(number);
        lcd.print("0.");
        lcd.print(num.substring(0, 3));
        lcd.print("m"); lcd.write(244);
        break;
      }
    case 2: {
        String num = String(number);
        lcd.print("0.0");
        lcd.print(num.substring(0, 2));
        lcd.print("m"); lcd.write(244);
        break;
      }
    case 1: {
        String num = String(number);
        lcd.print("0.00");
        lcd.print(num.substring(0, 1));
        lcd.print("m"); lcd.write(244);
        break;
      }
    default: {
        lcd.print("error("); lcd.print(d); lcd.print(")");
      }
  }
}

unsigned long roundIt(unsigned long what, unsigned int how) {
  return ((what + (how / 2)) / how) * how;
};

void loop() {
  oldNav = menu(mainMenu, mainMenuN, "Nabidka funkci", oldNav, 0);
  switch (oldNav) { //obsluha mainMenu
    case 0: { //NABIJ
        lcd.clear();
        lcd.print("Nabit kondenzator");
        lcd.setCursor(0, 1);
        lcd.print(" Proud: "); printValue(getParam(0), 0, 9, 1);
        lcd.setCursor(0, 2);
        lcd.print(" Napeti: "); printValue(getParam(1), 1, 9, 2);
        lcd.setCursor(0, 3);
        lcd.print(" ESC=Zpet "); lcd.write(6); lcd.print("=Spustit");
        do { //cekam na "spravnou" klavesu
          int key = getKey();
          if (key == 3) { //esc vyskoci
            break;
          } else if (key == 0) { //enter zacne
            int ret = nabij(1);
            if (ret == 0) {
              done(1, 3);
            } else if (ret == 1) {
              done(0, 1);
            } else if (ret == 2) {
              done(1, 0);
            }
            break;
          }
          //sipky ignoruj
        } while (1);
        break;
      }
    case 1: { //Vybij
        lcd.clear();
        lcd.print("Vybit kondenzator");
        lcd.setCursor(0, 1);
        lcd.print(" Proud: "); printValue(getParam(2), 2, 9, 1);
        lcd.setCursor(0, 2);
        lcd.print(" Napeti: "); printValue(getParam(3), 3, 9, 2);
        lcd.setCursor(0, 3);
        lcd.print(" ESC=Zpet "); lcd.write(6); lcd.print("=Spustit");
        do { //cekam na "spravnou" klavesu
          int key = getKey();
          if (key == 3) { //esc vyskoci
            break;
          } else if (key == 0) { //enter zacne
            int ret = vybij();
            if (ret == 0) {
              done(1, 2);
            } else if (ret == 1) {
              done(0, 1);
            } else if (ret == 2) {
              done(1, 0);
            }
            break;
          }
          //sipky ignoruj
        } while (1);
        break;
      }
    case 2: { //mereni C
        lcd.clear();
        lcd.print("Mereni kapacity");
        lcd.setCursor(0, 1);
        lcd.print(" Nab:"); printValue(getParam(0), 0, 11, 1);
        lcd.print("/");  printValue(getParam(1), 1, 16, 1);
        lcd.setCursor(0, 2);
        lcd.print(" Vyb:"); printValue(getParam(2), 2, 6, 2);
        lcd.print("/");  printValue(getParam(4), 4, 11, 2);
        lcd.print("/");  printValue(getParam(5), 5, 16, 2);
        lcd.setCursor(0, 3);
        lcd.print(" ESC=Zpet  "); lcd.write(6); lcd.print("=Spustit");
        do { //cekam na "spravnou" klavesu
          int key = getKey();
          if (key == 3) { //esc vyskoci
            break;
          } else if (key == 0) { //enter zacne
            lcd.setCursor(0, 1);
            lcd.write(165);
            int ret = nabij(1);
            if (ret == 0) {
              done(1, 3);
            } else if (ret == 2) {
              done(1, 0);
            } else if (ret == 1) { //nabiti ok, jedem dal
              lcd.setCursor(0, 1);
              lcd.write(" ");
              delay(3000);

              initProgressBar();
              int val; //hodnoty pro graf
              int usedI; //v mA
              int mez1 = (getParam(4) * 1023) / 25;
              int mez2 = (getParam(5) * 1023) / 25;

              lcd.setCursor(0, 2);
              lcd.write(165);
              switch (getParam(2)) {//zacni vybijet
                case 1: { //50mA
                    digitalWrite(5, LOW);
                    usedI = 5;
                    break;
                  }
                case 2: { //0.5A
                    digitalWrite(6, LOW);
                    usedI = 50;
                    break;
                  }
                case 3: { //5.0A
                    digitalWrite(7, LOW);
                    digitalWrite(8, HIGH); // bude zahul - bude vitr
                    usedI = 500;
                    break;
                  }
              }
              delay(100);

              int currU = analogRead(2);
              int firstU = currU;

              if (currU < mez1) { //vsechno spatne... pryc
                vypni();
                digitalWrite(8, LOW);
                done(1, 4);
                break;
              }

              boolean success = true;
              do { //PRVNI MEZ
                currU = analogRead(2);

                val = map(currU, firstU, mez2, 0, 60);
                updateProgress(val);

                if (esc) {
                  esc = false;
                  vypni();
                  digitalWrite(8, LOW); // bude zahul - bude vitr
                  done(1, 0);
                  success = false;
                  break;
                }
              } while (currU > mez1);
              if (!success) break;

              unsigned long t1 = millis(); //cas presahnuti meze
              unsigned long elapsedT = 0;

              do { //DRUHA MEZ
                currU = analogRead(2);

                val = map(currU, firstU, mez2, 0, 60);
                updateProgress(val);

                if (esc) {
                  esc = false;
                  vypni();
                  digitalWrite(8, LOW); // bude zahul - bude vitr
                  done(1, 0);
                  success = false;
                  break;
                }

                elapsedT = millis() - t1; //v ms
                if (elapsedT > 72000000) { //pokud cas >20h (72m ms)
                  vypni();
                  digitalWrite(8, LOW); // bude zahul - bude vitr
                  done(1, 5);
                  success = false;
                  break;
                }
              } while (currU > mez2);
              if (!success) break;

              vypni();
              digitalWrite(8, LOW); // bude zahul - bude vitr

              elapsedT /= 10;
              int deltaU = (getParam(4) - getParam(5));

              unsigned long result = usedI * elapsedT / deltaU;

              lcd.setCursor(0, 1);
              lcd.print("                    ");
              lcd.setCursor(0, 2);
              lcd.print("                    ");
              lcd.setCursor(1, 1);
              lcd.print("C = ");

              int d = (result == 0 ? 1 : ((int)(log10(result) + 1)));   
              switch (d) { //posledni cifra musi byt vzdy 0... pak se ta 0 umaze
                case 9: {
                    result = roundIt(result, 100000);
                    String num = String(result);
                    lcd.print(num.substring(0, 3));
                    lcd.print(".");
                    lcd.print(num.substring(3, 4));
                    lcd.print("kF");
                    break;
                  }
                case 8: {
                    result = roundIt(result, 10000);
                    String num = String(result);
                    lcd.print(num.substring(0, 2));
                    lcd.print(".");
                    lcd.print(num.substring(2, 4));
                    lcd.print("kF");
                    break;
                  }
                case 7: {
                    result = roundIt(result, 1000);
                    String num = String(result);
                    lcd.print(num.substring(0, 1));
                    lcd.print(".");
                    lcd.print(num.substring(1, 4));
                    lcd.print("kF");
                    break;
                  }
                case 6: {
                    result = roundIt(result, 100);
                    String num = String(result);
                    lcd.print(num.substring(0, 3));
                    lcd.print(".");
                    lcd.print(num.substring(3, 4));
                    lcd.print("F");
                    break;
                  }
                case 5: {
                    result = roundIt(result, 10);
                    String num = String(result);
                    lcd.print(num.substring(0, 2));
                    lcd.print(".");
                    lcd.print(num.substring(2, 4));
                    lcd.print("F");
                    break;
                  }
                case 4: {
                    String num = String(result);
                    lcd.print(num.substring(0, 1));
                    lcd.print(".");
                    lcd.print(num.substring(1, 4));
                    lcd.print("F");
                    break;
                  }
                case 3: {
                    String num = String(result);
                    lcd.print(num.substring(0, 3));
                    lcd.print("mF");
                    break;
                  }
                case 2: {
                    String num = String(result);
                    lcd.print(num.substring(0, 2));
                    lcd.print("mF");
                    break;
                  }
                case 1: {
                    String num = String(result);
                    lcd.print(num.substring(0, 1));
                    lcd.print("mF");
                    break;
                  }
                default: {
                    lcd.print("error("); lcd.print(d); lcd.print(")");
                  }
              }
              done(0, 1);
            }
          }
          break;
          //om nom nom sipky
        } while (1);
        break;
      }
    case 3: { //zmerit ESR
        lcd.clear();
        lcd.print("Mereni ESR");
        lcd.setCursor(0, 1);
        lcd.print(" Nab:"); printValue(getParam(0), 0, 6, 1);
        lcd.print("/");  printValue(getParam(1), 1, 11, 1);
        lcd.setCursor(0, 2);
        lcd.print(" Vyb:"); printValue(getParam(2), 2, 6, 2);
        lcd.print("/");  printValue(getParam(6), 6, 11, 2);
        lcd.print("/");  printValue(getParam(7), 7, 16, 2);
        lcd.setCursor(0, 3);
        lcd.print(" ESC=Zpet  "); lcd.write(6); lcd.print("=Spustit");
        do { //cekam na "spravnou" klavesu
          int key = getKey();
          if (key == 3) { //esc vyskoci
            break;
          } else if (key == 0) { //enter zacne
            lcd.setCursor(0, 1);
            lcd.write(165);
            int ret = nabij(1);
            if (ret == 0) {
              done(1, 3);
            } else if (ret == 2) {
              done(1, 0);
            } else if (ret == 1) {
              lcd.setCursor(0, 1);
              lcd.write(" ");
              delay(1000);
              lcd.setCursor(0, 2);
              lcd.write(165);

              ret = vybij(getParam(6));
              if (ret == 0) {
                done(1, 7);
                break;
              } else if (ret == 2) {
                done(1, 0);
                break;
              }
              
              // ESR1
              unsigned long esr1 = getESR();
              if (esr1 == 0) { //chyba
                done(1, 6);
                vypni();
                break;
              }

              bool failed = false;
              ret = vybij(getParam(7));
              if (ret == 0) {
                failed = true;
              } else if (ret == 2) {
                done(1, 0);
                break;
              }

              // ESR2
              unsigned long esr2 = 0;
              if (!failed) {
                esr2 = getESR();
                if (esr1 == 0) { //chyba
                  done(1, 6);
                  vypni();
                  break;
                }
              }

              vypni(); //jistota je jistota

              lcd.setCursor(0, 1);
              lcd.print("                    ");
              lcd.setCursor(0, 2);
              lcd.print("                    ");
              
              printValue(getParam(6), 6, 1, 1); lcd.print(": ");
              printESR(esr1);

              printValue(getParam(7), 7, 1, 2); lcd.print(": ");
              if(failed){
                lcd.print("--");
              } else {
                printESR(esr2);
              }

              done(0, 1);
            }
            break;
          }
          //sipky se co?... sipky se ignoruji
        } while (1);
        break;
      }
    case 4: { //Formovani
        lcd.clear();
        lcd.print("Formovani");
        lcd.setCursor(0, 1);
        lcd.print(" Nabiti: "); printValue(getParam(0), 0, 9, 1);
        lcd.print("/");  printValue(getParam(1), 1, 14, 1);
        lcd.setCursor(0, 2);
        lcd.print(" Cas: "); printValue(getParam(8), 8, 10, 2);
        lcd.setCursor(0, 3);
        lcd.print(" ESC=Zpet "); lcd.write(6); lcd.print("=Spustit");

        boolean cancel = false;
        do { //cekam na "spravnou" klavesu
          int key = getKey();
          if (key == 3) { //esc vyskoci
            break;
          } else if (key == 0) { //enter zacne
            lcd.setCursor(0, 1);
            lcd.write(165);
            int ret = nabij(1);
            if (ret == 0) {
              done(1, 3);
            } else if (ret == 2) {
              done(1, 0);
            } else if (ret == 1) {
              lcd.setCursor(0, 1);
              lcd.write(" ");
              lcd.setCursor(0, 2);
              lcd.write(165);
              unsigned long startT = millis()/100;
              unsigned long duration = getParam(8)*600;
              unsigned long elapsed = 0;
              int mez = ((getParam(1) * 1023) / 25) - 20;
              initProgressBar();              
              do {
                /////////
                /////////
                /////////
                elapsed = millis()/100-startT;
                int val = map(elapsed, 0, duration, 0, 60); 
                updateProgress(val);

                int currU = analogRead(2);

                if (currU < mez) {
                  lcd.setCursor(19, 0);
                  lcd.write(165);
                  if (nabij(0) == 2) {
                    cancel = true;
                    break;
                  }
                  lcd.setCursor(19, 0);
                  lcd.write(" ");
                }

                if (esc) {
                  esc = false;
                  cancel = true;
                  break;
                }
              } while (elapsed < duration);

              if (cancel) {
                done(1, 0);
              } else {
                lcd.setCursor(0, 2);
                lcd.write(" ");
                done(0, 1);
              }
            }
            break;
          }
          //sipky ignoruj
        } while (1);
        break;
      }
    case 5: { //Svod
        lcd.clear();
        lcd.print("Mereni svodu");
        lcd.setCursor(0, 1);
        lcd.print(" Nabiti: "); printValue(getParam(0), 0, 9, 1);
        lcd.print("/");  printValue(getParam(1), 1, 14, 1);
        lcd.setCursor(0, 2);
        lcd.print(" Cas: "); printValue(getParam(9), 9, 10, 2);
        lcd.print("/");  printValue(getParam(10), 10, 16, 2);
        lcd.setCursor(0, 3);
        lcd.print(" ESC=Zpet "); lcd.write(6); lcd.print("=Spustit");
        do { //cekam na "spravnou" klavesu
          int key = getKey();
          if (key == 3) { //esc vyskoci
            break;
          } else if (key == 0) { //enter zacne
            lcd.setCursor(0, 1);
            lcd.write(165);
            int ret = nabij(1);
            if (ret == 0) {
              done(1, 3);
            } else if (ret == 2) {
              done(1, 0);
            } else if (ret == 1) { //nabiti ok, jedu dal
              lcd.setCursor(0, 1);
              lcd.write(" ");
              lcd.setCursor(0, 2);
              lcd.write(165); //tecka u nabiti

              unsigned long startT = millis();
              unsigned long duration = getParam(9)*60000;
              unsigned long elapsed = 0;
              bool success = true;
              initProgressBar();
              do { //chvilu pockej - prodleva pred merenim
                elapsed = millis()-startT;
                int val = map(millis(), 0, duration, 0, 60);
                //zobraz(val);
                updateProgress(val);
                showVoltage();
                if (esc) {
                  esc = false;
                  done(1, 0);
                  success = false;
                  break;
                }
              } while (elapsed < duration);
              if(!success) break;

              //zmer poprve
              int u1 = 0;
              for (int i = 0; i < 20; i++) {
                u1 += analogRead(2);
                delay(50);
              }
              u1 /= 20;

              startT = millis();
              duration = getParam(10)*60000;
              elapsed = 0;
              success = true;
              initProgressBar();
              do { //cekame na uplynuti doby pro mereni
                elapsed = millis()-startT;
                int val = map(millis(), 0, duration, 0, 60);
                updateProgress(val);
                showVoltage();

                if (esc) {
                  esc = false;
                  done(1, 0);
                  success = false;
                  break;
                }
              } while (elapsed<duration);
              if(!success) break;

              //zmer podruhe
              int u2 = 0;
              for (int i = 0; i < 20; i++) {
                u2 += analogRead(2);
                delay(50);
              }
              u2 /= 20;

              lcd.setCursor(0, 2);
              lcd.write(" "); //tecka pryc

              int change = u1 - u2; //vypis rozdil
              lcd.setCursor(14, 0);
              lcd.print("      ");
              lcd.setCursor(0, 1);
              lcd.print(" Pokles:            ");
              lcd.setCursor(9, 1);
              lcd.print(change * 2.5 / 1023);
              lcd.print("V");
              lcd.setCursor(0, 2);
              lcd.print(" Doba:              ");
              printValue(getParam(10), 10, 10, 2);
              done(0, 1);
            }
            break;
          }
          //sipky ignoruj
        } while (1);
        break;
      }
    case 6: {
        int sub = menu(nastaveni, nastaveniN, "Nastaveni", 0, 1);
        if (sub == posOfSave) {
          lcd.clear();
          EEPROM.put(0, params);
          lcd.setCursor(1, 1);
          lcd.print("Ulozeno do EEPROM");
          delay(2000);
        }
        break;
      }
  }
  if (oldNav >= 100) {
    oldNav -= 100;
  }
}
